*  This is a demo MEMOEDIT() UDF which does WordPerfect style block cut/copy
*  functions.  It is written in pure Clipper code.  If you were to compile this 
*  program and link it with CLIPPER.LIB and EXTEND.LIB, you will see 
*  what it does.  Hit F3 to see a help menu.
*  This function uses a simple database file named MEMTEST.DBF
*  If this file does not exist, the program creates it.
*
*  The block functions have two major limitations;
*  1) You can only block the characters that are displayed on the MEMOEDIT()
*     window.
*  2) You must start blocking form the top and go down.
*  
*  I will be the first one to admit that the code could be written in a
*  much more efficient way (more procedure oriented code, etc.).  This code
*  is intended for illustrative purposes.  Please use it, rewrite it and 
*  if you have any comments, let me hear/read from you.
*
*  Programmer         : Semih Kangal 
*  CompuServe Acc. No.: 71211,3071


SET SCOREBOARD OFF
SET CONFIRM ON

*  Check to see if MEMTEST.DBF file exist.  If not, create it, and APPEND 
*  a blank record.
IF .NOT. FILE("MEMTEST.DBF")
    CREATE M_C_TEMP.MTM
    APPE BLANK
    REPLACE FIELD_NAME WITH "MEM_TEXT"
    REPLACE FIELD_TYPE WITH "M"
    REPLACE FIELD_LEN  WITH 10
    USE
    CREATE MEMTEST FROM M_C_TEMP.MTM
    APPE BLANK
    USE
    ERASE M_C_TEMP.MTM
ENDIF

*  Set default color variables for testing.
normal_color  = "W,I"
inverse_color = "I,W"
blank_color   = "W,W" 
lookup_color  = "I,W"
message_color = "I,W"

*  Call function to start maintaining color codes
sk_colcur="NT"              && N means normal_color, T means turn the cursor on

*  Open the test file and load the contents of the memo filed to variable TEXT
USE MEMTEST
text = MEM_TEXT

*  Initialize MEMOEDIT coordinates  (These variables are used by sk_memo()
br = 5                                  && Begining row for MEMOEDIT()  
bc = 3                                  && Beginnig col for MEMOEDIT()
er = 10                                 && Ending row for MEMOEDIT()
ec = 39                                 && Ending col for MEMOEDIT()
clp = ""                                && Clipboard
d_clp = ""                              && Temporary clipboard 
deleting = .F.                          && Deleting flag   
blk_beg_row = 0 
blk_beg_col = 0

*  Clear, draw screen and start MEMOEDIT()
CLEAR
@br-1,bc-1 TO er+3,ec+1 DOUBLE
@er+1,bc   TO er+1,ec   DOUBLE
text = MEMOEDIT(text,br,bc,er,ec,.T.,"sk_memo")

*  Replace the memo field and close the file
REPLACE MEM_TEXT with text
USE

*  End the procedure
@22,2 SAY "Well???"
@23,0

FUNCTION SK_MEMO
PARAMETERS m_stat,m_row,m_col 
PRIVATE key_pressed, ;         && LASTKEY()
        ik,          ;         && Temporary INKEY() storage for block
        temp_colcur, ;         && Temporary color variable for block
        blk_row,     ;         && Block row_position
        blk_col,     ;         && Block col_position 
        scr,         ;         && Temporary screen storage 
        key_res,     ;         && Keyboard response
        cntr,        ;         && Number of keystrokes after block was on
        dummy,       ;         && Temporary variable
        bchr                   && Dummy character variable 


IF deleting   && If the user wants to delete the block

   *  Get the character at the cursor position
   bchr=SUBSTR(SAVESCREEN(blk_beg_row,blk_beg_col,blk_beg_row, ;
               blk_beg_col),1,1)

   * If the temporary block variable is empty, or the character at the
   *    cursor position is not a blank .and. it is not contained in the
   *    temporary block variable, discontinue deleting.
   IF LEN(d_clp) = 0 .OR. (bchr <> " " .AND. AT(bchr,d_clp) = 0)
      deleting = .F.
      KEYBOARD CHR(2)
      RETURN 2
   ENDIF                        

   *  Strip out the first character from the temporary block variable.
   IF bchr =  SUBSTR(d_clp,1,1)
      d_clp = SUBSTR(d_clp,2,LEN(d_clp)-1)
   ENDIF   

   * Send a delete keystroke to MEMOEDIT()
   KEYBOARD CHR(7)
   RETURN 2

ENDIF    && End of 'IF deleting'

DO CASE
   CASE m_stat = 0  && When in idle, reformat memo field.
        RETURN 2
   CASE m_stat = 3  && Initialization; Set insert on if it is not on.
        IF .NOT. READINSERT()
           RETURN 22
        ENDIF
        @er+2,bc+1 SAY "Insert   "
        RETURN 0
   OTHERWISE        && A re-configurable key was pressed !!  Key exception
        key_pressed = LASTKEY()
        DO CASE


           CASE key_pressed = 22       && Insert toggle
              IF READINSERT()
                 @er+2,bc+1 SAY "Overwrite"
              ELSE
                 @er+2,bc+1 SAY "Insert   "
              ENDIF
              RETURN 22


           CASE key_pressed = 27 .OR. key_pressed = -6  && F7 or Esc (Exit)
              IF m_stat = 2       && Memo was changed
                 sk_sound(2,3)                   && Sound the tone
                 key_res = "N"
                 SET CONFIRM OFF
                 key_res = sk_memo_input("Abandon changes? (Y/N):", ;
                           key_res,"YN")
                 SET CONFIRM ON
                 IF key_res = "Y"       && Abondon edit
                    RETURN 27
                 ELSE
                    RETURN 32           && Ignore the previous key stroke
                 ENDIF
              ELSE
                 RETURN 27
              ENDIF


           CASE key_pressed = -2       && F3 (Help)
              DO sk_memohelp


           CASE key_pressed = -3       && F4 (Block)

              *  Set the clipboard variable to EMPTY
              clp = ""

              *  Get the block_beginning_coordinates 
              blk_beg_row = ROW() 
              blk_beg_col = COL() 

              *  Initialize the block_tracking variables
              blk_row = blk_beg_row
              blk_col = blk_beg_col

              *  Turn on the 'BLOCK' sign
              temp_colcur = sk_colcur
              SET COLOR TO W*+
              @er+2,ec-5 say "Block "
              sk_colcur(temp_colcur)

              *  Put the cursor at the block_beginning_location
              @blk_row,blk_col SAY ""
              *  Initialize the keyboard counter
              cntr = 0
              ik = 0
 
              *  Start monitoring the keyboard
              DO WHILE ik <> -3
                ik = INKEY(0)
                DO CASE
                   CASE ik = 5       && UP ARROW
                        * Stuff keyboard with Left_arrow characters
                        KEYBOARD REPLICATE(CHR(26),ec-bc+1)
                        LOOP
                   CASE ik = 24      && DOWN ARROW
                        * Stuff keyboard with Right_Arrow characters
                        KEYBOARD REPLICATE(CHR(4),ec-bc+1)
                        LOOP
                   CASE ik = 4       && RIGHT ARROW
                        *  If cursor position is at the end of the MEMOEDIT()
                        *     window, flush the keyboard and sound the tone.    
                        IF blk_col > ec
                           sk_sound(2,3)
                           KEYBOARD ""
                           LOOP
                        ENDIF
                        *  Get the character at the cursor location
                        bchr = substr(SAVESCREEN(blk_row,blk_col,blk_row, ;
                                       blk_col),1,1)
                        *  Add the character to the clipboard
                        clp = clp + bchr
                        *  Display the character in inverse color
                        temp_colcur = sk_colcur("I ")
                        @blk_row,blk_col SAY bchr 
                        sk_colcur(temp_colcur)
                        * Add 1 to the keyboard counter and cursor_location_
                        *     tracer. 
                        cntr = cntr + 1
                        blk_col = blk_col + 1  
                        *  If the cursor is at the end of the MEMOEDIT() 
                        *     line, see if you can go 1 line down.
                        IF blk_col > ec 
                           IF blk_row + 1 <= er
                              * Move down one line 
                              blk_row = blk_row + 1
                              blk_col = bc
                              @blk_row,blk_col SAY ""
                           ELSE
                              *  You are at the lower right corner of the 
                              *  MEMOEDIT() window.  Sound the tone and
                              * do not advance the cursor.
                              sk_sound(2,3)
                              @blk_row,blk_col-1 SAY "" 
                           ENDIF
                        ENDIF
                        LOOP
                   CASE ik = 19 .OR. ik = 26   && LEFT ARROW or Ctrl-Leftarrow
                        IF cntr > 0    && If the cursor is left and/or 
                                       ** under the beginning of block
                           ** A
                           IF LEN(clp) = 1
                              clp = ""
                           ELSE
                              clp = SUBSTR(clp,1,LEN(clp)-1)
                           ENDIF
                           ** B
                           blk_col = blk_col - 1
                           DO CASE
                              CASE blk_col > bc    && If it is the same line
                                   @blk_row,blk_col SAY ;
                                         substr(SAVESCREEN(blk_row,blk_col,; 
                                         blk_row,blk_col),1,1)
                              CASE blk_col = bc    && If it is the same line
                                   @blk_row,blk_col SAY ;
                                         substr(SAVESCREEN(blk_row,blk_col,;
                                         blk_row,blk_col),1,1)
                                   @blk_row,blk_col-1 SAY ;
                                         substr(SAVESCREEN(blk_row,blk_col-1,;
                                         blk_row,blk_col-1),1,1)
                              CASE blk_col < bc    && If you must go up
                                   @blk_row-1,ec  SAY ;
                                         substr(SAVESCREEN(blk_row-1,ec,;
                                         blk_row-1,ec),1,1)
                           ENDCASE 
                            ** C
                           IF cntr > 1
                              temp_colcur = sk_colcur("I ")
                           ENDIF
                           DO CASE
                              CASE blk_col < bc
                                   blk_col = ec
                                   blk_row = blk_row - 1
                                   @blk_row,blk_col SAY ""
                              CASE blk_col > bc
                                   @blk_row,blk_col SAY ""
                           ENDCASE
                           IF cntr > 1
                              sk_colcur(temp_colcur) 
                           ENDIF 
                           ** D
                           cntr = cntr - 1

                        ELSE  && If the cursor is right and/or above the B.O.B.
                              sk_sound(2,3)
                              KEYBOARD ""
                        ENDIF
                        LOOP

                   CASE ik = -3      && End of block.

                   CASE ik = 7       && Delete block
                        * User wants to delete the block.  Copy the clipboard
                        * to a temporary clipboard.  Turn deleting flag on,
                        * erase the 'BLOCK' sign, and return to MEMOEDIT()
                        deleting = .T.
                        d_clp = clp
                        @er+2,ec-5 say "      "
                        KEYBOARD CHR(2)
                        RETURN 2

                   CASE ik = -24     && Save  Ctrl-F5
                        *  Strip out the double blanks out of the block. 
                        dummy = 1
                        DO WHILE dummy > 0
                           dummy = AT("  ",clp)
                           IF dummy > 0
                              clp = LEFT(clp,dummy) + SUBSTR(clp,dummy+2)
                           ENDIF
                        ENDDO 
                        * Get the filename to write the block to  
                        key_res = SPACE(12)
                        key_res = sk_memo_input("Save block to:",key_res," ")
                        IF EMPTY(key_res) .OR. LASTKEY() = 27
                           LOOP
                        ENDIF
                        key_res = ALLTRIM(key_res)
                        *  If the file exists, do the usual.
                        IF FILE(key_res)
                           SET CONFIRM OFF
                           IF sk_memo_input("Replace "+UPPER(key_res)+"? (YN)", ;
                              "N","YN") = "N"
                               @blk_row,blk_col SAY ""
                           ELSE
                              ik = -3
                              MEMOWRIT(key_res,clp)
                              sk_memo_wait("File "+key_res+" is written!")
                              KEYBOARD CHR(13)
                           ENDIF
                           SET CONFIRM ON
                        ELSE
                           ik = -3
                           MEMOWRIT(key_res,clp)
                           sk_memo_wait("File "+key_res+" is written!")
                           KEYBOARD CHR(13)
                        ENDIF


                   OTHERWISE
                        sk_sound(1,1)
                 ENDCASE
              ENDDO

              *  OK, the user turned the block off
              @er+2,ec-5 say "      "
              * If the block is empty, do nothing
              IF cntr = 0
                 RETURN 2
              ENDIF

              *  Strip double blanks out of the clipboard.
              dummy = 1
              DO WHILE dummy > 0
                 dummy = AT("  ",clp)
                 IF dummy > 0
                    clp = LEFT(clp,dummy) + SUBSTR(clp,dummy+2)
                 ENDIF
              ENDDO 

              *  What do you want to do?
              key_res = "0"
              SET CONFIRM OFF
              key_res = sk_memo_input("1)Copy 2)Move :",key_res, ;
                    "012")
              SET CONFIRM ON
              * If 'Move', activate the deleting
              IF key_res = "2"
                 deleting = .T.
                 d_clp = clp
                 KEYBOARD CHR(2)
              ENDIF
              RETURN 2


           CASE key_pressed = -9     && F10
                *  F10; Save the contents and quit MEMOEDIT()
                KEYBOARD CHR(2)+CHR(23)
                save_it = .T.
                RETURN 2


           CASE key_pressed = -24    && Ctrl-F5
                *  Do you want to save the memofield, or retrieve a text file
                *  into the memo field?
                key_res = "0"
                SET CONFIRM OFF
                key_res = sk_memo_input("1)Save 2)Retrieve :",key_res, ;
                    "012")
                SET CONFIRM ON
                DO CASE
                   CASE key_res = "1"   && Save
                        IF m_stat = 2   && Memo was changed
                           key_res = "Y"
                           SET CONFIRM OFF
                           key_res = sk_memo_input("Changes won't be"+ ;
                                     " saved, OK? (Y/N):",key_res,"YN")
                           SET CONFIRM ON
                           IF LASTKEY() = 27 .OR. key_res = "N"
                              RETURN 2
                           ENDIF
                        ENDIF 
                        key_res = SPACE(12)
                        key_res = sk_memo_input("Save to:",key_res," ")
                        IF EMPTY(key_res) .OR. LASTKEY() = 27 .OR. ;
                                               LASTKEY() = -6
                           RETURN 2
                        ENDIF
                        *  If the file already exists, do the ususal.
                        key_res = ALLTRIM(key_res)
                        IF FILE(key_res)
                           SET CONFIRM OFF
                           IF sk_memo_input("Replace "+UPPER(key_res)+;
                                     "? (YN)","N","YN") <> "N"
                              MEMOWRIT(key_res,text)
                              sk_memo_wait("File "+key_res+" is written!")
                           ENDIF
                           SET CONFIRM ON 
                        ELSE
                           MEMOWRIT(key_res,text)
                           sk_memo_wait("File "+key_res+" is written!")
                        ENDIF
                        RETURN 2
                   CASE key_res = "2"   && Retrieve
                        key_res = SPACE(12) 
                        IF .NOT. EMPTY(clp)
                            key_res = "(BLOCK)     "
                        ENDIF
                        key_res = sk_memo_input("Retrieve:",key_res," ")
                        IF EMPTY(key_res) .OR. LASTKEY() = 27
                           RETURN 0
                        ENDIF
                        key_res = ALLTRIM(key_res)
                        DO CASE
                           CASE key_res = "(BLOCK)"
                                KEYBOARD clp+CHR(2)
                           OTHERWISE
                                IF FILE(key_res)
                                   KEYBOARD MEMOREAD(key_res)+CHR(2)
                                ELSE
                                   sk_memo_wait("File not found!")
                                ENDIF
                        ENDCASE
                ENDCASE
        ENDCASE
ENDCASE
RETURN 0

FUNCTION sk_memo_input
*  This function does a @.. SAY  GET within the MEMOEDIT() window coordinates
PARAMETERS wht_msg, wht_read, wht_valid
PRIVATE scr,temp_colcur
scr = SAVESCREEN(er+2,bc,er+2,ec)
temp_colcur = sk_colcur("B ")
@er+2,bc CLEAR TO er+2,ec
IF .NOT. EMPTY(wht_valid)
   @er+2,bc SAY wht_msg GET wht_read PICTURE REPLICATE("!",LEN(wht_read)) ;
                        VALID (wht_read $ wht_valid)
ELSE
   @er+2,bc SAY wht_msg GET wht_read PICTURE "@K "+REPLICATE("!",LEN(wht_read))
ENDIF
READ      
sk_colcur(temp_colcur)
RESTSCREEN(er+2,bc,er+2,ec,scr)
RETURN (wht_read)

FUNCTION sk_memo_wait
*  This function displays a message within the MEMOEDIT() window coordinates
PARAMETERS wht_msg
PRIVATE scr,temp_colcur
scr = SAVESCREEN(er+2,bc,er+2,ec)
temp_colcur = sk_colcur("B ")
@er+2,bc CLEAR TO er+2,ec
@er+2,bc SAY wht_msg
sk_sound(2,2)
inkey(1)
sk_colcur(temp_colcur)
RESTSCREEN(er+2,bc,er+2,ec,scr)
RETURN (.T.)

PROCEDURE sk_memohelp
*  A simple sk_memo() help function for test purposes
PRIVATE SCR,r,c,oldcolcur
SAVE SCREEN TO SCR
oldcolcur = sk_colcur("IF")
r=10
c=10
@r,c clear to r+14,c+65
@r,c       to r+14,c+65 double
@r+12,c+1  to r+12,c+64 double
@R+1,C+1 SAY "                    H E L P                                    "
@R+3,C+1 SAY " F3        =  Help                                             "
@R+4,C+1 SAY " F4        =  Turn on/off block                                "
@R+5,C+1 SAY " Ctrl-F5   =  Text in/out (Save/Retrieve memo field or         "
@R+6,C+1 SAY "                           save the block)                     " 
@R+7,C+1 SAY " F7        =  Exit                                             "
@R+8,C+1 SAY " F10       =  Save & Exit                                      "
@R+9,C+1 SAY " Del       =  Delete the whole block.                          " 

@r+13,c+1 SAY "        Hit any key to return to program ....................  "
inkey(0)
sk_colcur(oldcolcur)
restore screen from scr
RETURN

PROCEDURE sk_sound
* This function sounds the bell.  1-Low C, 2-Mid C, 3-High C
* 1,2,3,4,5 ....  duration
PARAMETERS sk_octave,sk_duration
PRIVATE sk_pcount
sk_pcount = PCOUNT()
DO CASE
   CASE sk_pcount = 0
        sk_octave = 1
        sk_duration = 3   
   CASE sk_pcount = 1
        sk_duration = 3   
ENDCASE            
sk_octave = IF(sk_octave > 0 .AND. sk_octave < 4, sk_octave, 3)
sk_duration = IF(sk_duration < 1, 2, sk_duration)
DO CASE
   CASE sk_octave = 1
        TONE(130.80,sk_duration)
   CASE sk_octave = 2
        TONE(261.70,sk_duration)
   CASE sk_octave = 3
        TONE(523.30,sk_duration)
ENDCASE
RETURN 

FUNCTION sk_colcur
*  This function returns the logical current value of the color and the cursor
*  status.  The first character of the parameter is the color, the second is
*  the cursor.   
*  Color values: N=normal,I=inverse,B=blank,L=lookup,M=message,' '=ignore.
*                normal_color,inverse_color,blank_color,message_color   
*  Cursor values: T=on,F=off,' '=ignore
PARAMETER newcolcur
PRIVATE newcolor, newcursor, retcolcur
IF TYPE("sk_color_mode") = "U"
   PUBLIC sk_color_mode 
   sk_color_mode = "N"
ENDIF
IF TYPE("sk_cursor_mode") = "U"
   PUBLIC sk_cursor_mode 
   sk_cursor_mode = "T"
ENDIF
retcolcur = sk_color_mode + sk_cursor_mode
IF PCOUNT() <> 0
   newcolor = SUBSTR(newcolcur,1,1)
   newcursor= SUBSTR(newcolcur,2,1)
   DO CASE
      CASE newcolor = "N"
           SET COLOR TO (normal_color)
           sk_color_mode = "N"
      CASE newcolor = "I"
           SET COLOR TO (inverse_color)
           sk_color_mode = "I"
      CASE newcolor = "B"
           SET COLOR TO (blank_color)
           sk_color_mode = "B"
      CASE newcolor = "L"
           SET COLOR TO (lookup_color)
           sk_color_mode = "L"
      CASE newcolor = "M"
           SET COLOR TO (message_color)
           sk_color_mode = "M"
   ENDCASE
   DO CASE
      CASE newcursor = "T"
           SET CURSOR ON
           sk_cursor_mode = "T"
      CASE newcursor = "F"
           SET CURSOR OFF
           sk_cursor_mode = "F"
   ENDCASE
ENDIF
RETURN (retcolcur)

